/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

/* Basic definitions of games and operations on games, as in Winning
 * Ways (Berlekamp, Conway, Guy).  The functions for adding, negating
 * and comparing games are given here.  Most functions are "memoized",
 * meaning they first check if they have been called before with
 * the same arguments before executing.  Note that once you do a
 * make (left, right), both lists are no longer guaranteed to exist,
 * although the lists will be freed properly.  This is because
 * usually once you do a make, you typically only need to manipulate
 * the resulting game.
 */

/* Be careful about the list allocation and deallocation!  A good
 * sample function to look at is plus in gameops.c.  Note in
 * particular the use of list_unordered_2 and list_copy_unordered_2
 * for hash table lookups.  list_unordered_2 is used ONLY for
 * temporary reference.  hash_put() needs to be given a list
 * it can keep.  In summary, the function which deallocate
 * a list automatically are list_1, list_2, ... (which never
 * really allocate a list to begin with), make, hash_put, vs,
 * and list_clobber.  Otherwise you should do a list_free.
 */

#define GAME_BLOCK_SIZE 1000
#define NO_GAME ((game_type) 0)
#define CANONICAL_FLAG	(1<<0)
#define EVALUATED_FLAG	(1<<1)

typedef list_type var_type;

typedef struct game_struct {
  list_type left;
  list_type right;
  Q_type Num;
  short int Up;
  short int Star;
  unsigned char flags;
  var_type var;			/* Variable pointing to g set by set_var */
  struct game_struct *next;
#ifdef DEBUG
  int reference_count;		/* for debugging */
#endif
} *game_cell_type;

extern game_type	zero;
extern game_type	one;
extern game_type	two;
extern game_type	neg1;
extern game_type	one_star;
extern game_type	saltus;
extern int		period;

extern void		init(void);
extern void		reinit(void);
extern void		clean(void);
extern game_type	make (list_type, list_type);
extern game_type	make_canonical (list_type, list_type);
extern game_type	make_nonpartisan (list_type);
extern list_type	left_options (game_type);
extern list_type	right_options (game_type);
/* #define's
extern unsigned		count_left_options (game_type);
extern unsigned		count_right_options (game_type);
extern game_type	nth_left_option (game_type, unsigned);
extern game_type	nth_right_option (game_type, unsigned);

extern game_type	g_int (int);
extern game_type	num (Q_type);
extern game_type	star (int);
extern game_type	up (int);
extern game_type	down (int);
extern game_type	tiny (game_type);
extern game_type	miny (game_type);
*/
extern game_type	num_up_star (Q_type, short, short);

extern boolean		ge (game_type, game_type);	/* greater or equal */
extern boolean		le (game_type, game_type);	/* less than or equal */
extern boolean		eq (game_type, game_type);
extern boolean		gt (game_type, game_type);
extern boolean		lt (game_type, game_type);
extern char *		compare (game_type, game_type);	/* returns ">", "<", "=", or "<>" */
extern void             remove_dominated_options(list_type, test_type);

extern game_type	plus (game_type, game_type);
extern game_type	negative (game_type);
extern game_type	minus (game_type, game_type);
extern game_type	times (int, game_type);

extern game_type	cool (game_type, Q_type);
extern game_type	project (game_type);      /* Simplest game infinetsimally near g */
extern Q_type		temperature (game_type);
extern game_type	freeze (game_type);
extern game_type	overheat (game_type, game_type, game_type); /* Berlerkam's Blockbusting paper */
extern game_type	heat (game_type, game_type);
extern game_type	mult (game_type, game_type); /* Norton multiply */
extern Q_type		mean (game_type);
extern game_type	atomic_weight (game_type);
extern game_type	on (game_type, int);       /* Type "help ons" */
extern game_type	off (game_type, int);      /* Type "help ons" */
extern game_type	gexp(game_type, int);      /* Type "help ons" */
extern game_type	zeros_vs (game_type, int); /* 0|||0||0 ... |g */
extern game_type	vs_zeros (game_type, int);
extern game_type	vs (game_type, game_type); /* g|h */

extern Q_type		left_stop (game_type);
extern Q_type		right_stop (game_type);
extern list_type	incentives (game_type);		/* Union of Left's incentives with Right's */
extern list_type	left_incentives (game_type);
extern list_type	right_incentives (game_type);

extern boolean		is_int (game_type);
extern boolean		is_num (game_type);
extern boolean		is_up (game_type);
extern boolean		is_star (game_type);
extern boolean		is_tiny (game_type);
extern boolean		is_miny (game_type);
extern boolean		is_on (game_type);
extern boolean		is_num_up_star (game_type);
extern boolean		is_num_star (game_type);
extern boolean		is_up_star (game_type);
extern char *		var_sprintn (char *, var_type, int);
extern boolean		is_var (game_type);

/* #define's
extern int		int_value (game_type);
extern Q_type		num_value (game_type);
extern int		star_value (game_type);
*/
extern int		up_value (game_type);
extern game_type	tiny_value (game_type);
extern game_type	miny_value (game_type);
extern int		on_exponent (game_type);
extern game_type	on_value (game_type);
extern var_type		var_value (game_type);		/* Return variable assigned to game */
extern char *		var_sprint (char *, var_type);

/* Variables names are encoded in lists to make use of hash routines */
extern var_type		make_var (char *);
extern void		set_var (var_type, game_type);
extern game_type	get_var (var_type);

extern game_cell_type	game_cell[];	/* shouldn't really be externed -- for debugging */
extern char		*parser_input;
extern game_type	parser_output;
extern var_type		dummy_var;

#define _flags(g) (game_cell[g]->flags)
#define extended_left_options(g) (game_cell[g]->left)
#define extended_right_options(g) (game_cell[g]->right)
#define count_left_options(g) list_length (left_options (g))
#define count_right_options(g) list_length (right_options (g))
#define count_extended_left_options(g) list_length (extended_left_options (g))
#define count_extended_right_options(g) list_length (extended_right_options (g))
#define nth_left_option(g,i) list_nth (left_options (g), i)
#define nth_right_option(g,i) list_nth (right_options (g), i)

#define num_value(g) (game_cell[g]->Num)
#define int_value(g) Q_p (num_value (g))
#define star_value(g) (game_cell[g]->Star)
#define up_value(g) (game_cell[g]->Up)

#define num(n) num_up_star ((n), 0, 0)
#define g_int(n) num (Q_int (n))
#define star(n) num_up_star (Q_0, 0, (n))
#define up(n) num_up_star (Q_0, (n), 0)
#define down(n) num_up_star (Q_0, -(n), 0)
#define tiny(g) vs (zero, vs (zero, negative (g)))
#define miny(g) vs (vs ((g), zero), zero)
#define set_flag(g,f) (game_cell[g]->flags |= (f))
#define clear_flag(g,f) (game_cell[g]->flags &= ~(f))
