/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

/* Maintains a SORTED list of elements.  Done with abstract data types,
 * except functions list_elt_less (e1, e2) and list_elt_equal(e1,e2)
 * compare 2 elements, list_elt_copy(e1,e2) copies elt e2 into e1,
 * and list_elt_small makes an element smaller than all other elements.
 * Lastly, list_combine is very implementation dependent for a cluge.
 */

#define LIST_CELL_BLOCK_SIZE 1000	/* size of mallocs, when needed */
#define list_elt_equal(x,y) (x==y)
#define list_elt_less(x,y) (x<y)
#define list_elt_copy(x,y) (x=y)
#define list_elt_small(x) (x = 0)

typedef int list_elt_type;
typedef struct list_struct {
	list_elt_type elt;
	struct list_struct *next;
} *list_type;

#define NIL (list_type) 0

#if 0
/* There are the politically correct definitions, since the objects
 * involved are really list_elt_type and not game_type.  But the
 * only functions that call them pass functions of type
 *
 * game_type (*f)(game_type, game_type),
 *
 * and since unsigned and int (game_type and list_elt_type) are
 * guaranteed to be the same size, there is no harm done.
 *
 * One must always worry about ones-complement and sign-magnitude machines,
 * but since we don't do comparisons or arithmetic (just assignment),
 * we're okay.
 *
 */
typedef list_elt_type (*function_two_type)(list_elt_type, list_elt_type);
typedef list_elt_type (*function_one_type)(list_elt_type);
typedef boolean (*predicate_type)(list_elt_type);
#else
/* The politically incorrect but expedient declarations. */
typedef game_type (*function_two_type)(game_type, game_type);
typedef game_type (*function_one_type)(game_type);
typedef boolean (*predicate_type)(game_type);
#endif

extern long unsigned	list_count;	/* Number of lists allocated, not on free stack */

extern void		list_init(void);
extern list_type	list_make(void);
/* list_1, list_2, list_3,  and list_unordered_2 are functions returning
 * a recycled list of the indicated length.  Note that the list gets trashed the
 * next time the functions are called.  Used for temporary, fast, access. */
extern list_type	list_0;
extern list_type	list_1 (list_elt_type);
extern list_type	list_2 (list_elt_type, list_elt_type);
extern list_type	list_unordered_2 (list_elt_type, list_elt_type);
extern list_type	list_3 (list_elt_type, list_elt_type, list_elt_type);
extern void		list_insert (list_type, list_elt_type);
extern void		list_prepend (list_type, list_elt_type);
/* for backwards compatability */
#define list_insert_unsorted list_prepend
extern boolean		list_member (list_type, list_elt_type);
extern int		list_search (list_type, list_elt_type);
extern int		list_length (list_type);
extern void		list_delete (list_type, list_elt_type);
extern void		list_delete_nth (list_type, int);
extern void		list_change_nth_to (list_type, int, int);
extern list_elt_type	list_first  (list_type);
extern list_elt_type	list_nth (list_type, int);
extern void		list_free (list_type);
extern boolean		list_equal (list_type, list_type);
extern list_type	list_copy (list_type);
extern void		list_clobber (list_type, list_type);
		/* l1 <- l2 contents, clobbering both l1 and l2 old contents */
extern list_type	list_combine (list_type, list_type);

extern list_type	list_map (function_one_type, list_type);
	/* mapll acts on corresponding pairs, mapcomb acts on all pairs */
extern list_type	list_mapll (function_two_type, list_type, list_type);
extern list_type	list_mapcomb (function_two_type, list_type, list_type);
extern list_type	list_maplc (function_two_type, list_type, list_elt_type);
extern list_type	list_mapcl (function_two_type, list_elt_type, list_type);
extern boolean		list_every (predicate_type, list_type);
extern list_type	list_union (list_type, list_type);
extern list_type	list_multi_union (list_type, list_type);  /* allows for repeats */

extern unsigned long	list_stat (void);

#define list_copy_1(a) list_copy (list_1 (a))
#define list_copy_2(a,b) list_copy (list_2 (a,b))
#define list_copy_unordered_2(a,b) list_copy (list_unordered_2 (a,b))
#define list_copy_3(a,b,c) list_copy (list_3 (a,b,c))

/* list_start, list_rest and list_pick are used in tandem to step through
 * a list.  list_start() returns the beginnging of a list, and list_rest
 * returns successive elements in the list.  A return of NIL indicates end
 * of list.  Calling list_pick() return the actual element.
 * for (l = list_start (list); l != NIL; l = list_rest (l)) play_with (list_pick (l));
 */
#define list_start(l) (l->next)
#define list_rest(l) (l->next)
#define list_pick(l) (l->elt)
