/* ====================================================================== *\
   TOADS.C -- Play Toads and Frogs.  For rules, see _Winning Ways_, p.14.

   Copyright (C) Jeff Erickson, 1993.  All rights reserved.
   Permission is given to copy this file as part of David Wolfe's
   games program, provided nobody makes any money in the process.
\* ====================================================================== */

#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include <ctype.h>
#include "externs.h"
#include "list.h"
#include "hash.h"
#include "rational.h"
#include "gameops.h"
#include "output.h"
#include "bitmatrx.h"
#include "toads.h"

/* ---------------------------------------------------------------------- *\
|  void toads_trace(T_type pos, game_type g, char *msg, int level)
|
|  Trace output.  Does nothing unless TRACE_FLAG is set.  Output is
|  inset 'level' times. 
\* ---------------------------------------------------------------------- */
void toads_trace(T_type pos, game_type g, char *msg, int level)
{
    int i;

    if (!(flags&TRACE_FLAG))
	return;
    if ((tlevel != 0) && (level > tlevel))
	return;

    for (i=0; i<level; i++)
	putchar('|'); 
    putchar(' ');

    putchar('[');
    for (i=0; i<pos->length; i++)
    {
	switch (pos->board[i+2])
	{
	    case Blank:	putchar(' '); break;
	    case Frog:  putchar('f'); break;
	    case Toad:  putchar('t'); break;
	    case Wall:  putchar('|'); break;
	    default:	putchar('?'); break;
	}
    }
    putchar(']');
    if (pos->ttffbonus > 0)
	printf(" + %d", pos->ttffbonus);
    else if (pos->ttffbonus < 0)
	printf(" - %d", -pos->ttffbonus);

    if (g != NO_GAME)
    {
	printf(" = ");
	game_print(g);
    }

    if (msg)
	printf("    [%s]", msg);

    putchar('\n');
}

/* ----------------------------------------------------------------- *\
|  Print a position and its value.
\* ----------------------------------------------------------------- */
void T_print(T_type pos, game_type g, char *msg)
{
    int i;

    putchar('[');
    for (i=0; i<pos->length; i++)
    {
	switch (pos->board[i+2])
	{
	    case Blank:	putchar('-'); break;
	    case Frog:  putchar('F'); break;
	    case Toad:  putchar('T'); break;
	    case Wall:  putchar('|'); break;
	    default:	putchar('?'); break;
	}
    }
    putchar(']');
    if (pos->ttffbonus > 0)
	printf(" + %d", pos->ttffbonus);
    else if (pos->ttffbonus < 0)
	printf(" - %d", -pos->ttffbonus);

    if (g != NO_GAME)
    {
	printf(" = ");
	game_print(g);
    }

    if (msg)
	printf("    [%s]", msg);

    putchar('\n');
}


/* ---------------------------------------------------------------------- *\
|  T_type T_new(int length)
|
|  Get a new toads position with <length> slots
\* ---------------------------------------------------------------------- */
T_type T_new(int length)
{
    T_type t;

    disable_interrupts;
    t = (T_type) malloc(sizeof(struct T));
    t->length = length;
    t->components = 0;
    t->ttffbonus = 0;
    t->board = (T_Item *) malloc((length+5) * sizeof(T_Item));
    t->board[0] = t->board[1] = Wall;
    t->board[length+2] = t->board[length+3] = Wall;
    enable_interrupts;
    return t;
}

/* ---------------------------------------------------------------------- *\
|  void T_free(T_type pos)
|
|  Free a toads position
\* ---------------------------------------------------------------------- */
#define T_free(pos) if(pos){free((pos)->board);free(pos);}

/* ---------------------------------------------------------------------- *\
|  list_type T_to_list(T_type pos)
|
|  Convert a T_type to a list.  [Wolfe should have called them "sets"
|  instead of "lists", since insertion order doesn't matter!]
\* ---------------------------------------------------------------------- */
list_type T_to_list(T_type pos)
{
    list_type l;
    int i;

    l = list_make();
    for (i=0; i<pos->length; i++)
	list_insert(l, 10*i + (int) pos->board[i+2]);
    return l;
}


/* ---------------------------------------------------------------------- *\
|  T_type input_toads(void)
|
|  Ask the user to input a Toads and Frogs game and deal with it.
\* ---------------------------------------------------------------------- */
T_type input_toads(void)
{
    T_type pos;
    char s[MAXBUFF];
    int i;
    
    printf("Enter position (t=toad, f=frog, |=wall):\n");
    input_line(s);

    pos = T_new(strlen(s));
    pos->components = 1;

    for (i=0; s[i]; i++)
    {
	switch (tolower(s[i]))
	{
	    case 't':
	    case 'l':
	    case '>':
	        pos->board[i+2] = Toad;
		break;

	    case 'f':
	    case 'r':
	    case '<':
		pos->board[i+2] = Frog;
		break;

	    case '|':
	    case '#':
		pos->board[i+2] = Wall;
		pos->components++;
		break;

	    default:
		pos->board[i+2] = Blank;
		break;
	}
    }

    return pos;
}


/* ---------------------------------------------------------------------- *\
|  game_type toads_component(T_type pos, int level)
|
|  Evaluate a simplified single-component toads and frogs position
|  with no ttff bonus, and memoize the result.  Check for the Death
|  Leap Principle:    
|
|	Any position in which the onlylegal moves are jumps into
|	single spaces has value zero.
|
|  The positions to which the DLP applies form a regular expression,
|  so we can check for them using a finite state machine!
\* ---------------------------------------------------------------------- */
game_type toads_component(T_type pos, int level)
{
    typedef enum {
	Start=0,
	Got_, Got__, Got__T, Got_T, Got_F,
	GotT, GotT_, GotTF, GotTF_,
	GotF, GotF_,
	Die, Live
    } State;

    static State nextstate[12][4] =
    {
    /*  Wall	Toad	Frog	Blank 	*/	/* State x T_Item -> State */

        {Die,	GotT,	Start,	Got_	},	/* Start  */
        {Die,	Got_T,	Got_F,	Got__	},	/* Got_   */
        {Die,	Got__T,	Live,	Got__	},	/* Got__  */
        {Die,	GotT,	Live,	Live	},	/* Got__T */
        {Die,	Got_T,	GotTF,	Live	},	/* Got_T  */
        {Die,	GotT,	Got_F,	Live	},	/* Got_F  */
        {Die,	GotT,	GotTF,	GotT_	},	/* GotT   */
        {Die,	Live,	Live,	Live	},	/* GotT_  */
        {Die,	GotT,	GotF,	GotTF_	},	/* GotTF  */
        {Die,	Got_T,	Live,	Live	},	/* GotTF_ */
        {Die,	GotT,	GotF,	GotF_	},	/* GotF   */
        {Die,	Got_T,	Live,	Got__	},	/* GotF_  */
    };

    list_type left, right;
    game_type g = zero;
    list_type l;
    int i, moves;
    State thestate;

    toads_trace(pos, NO_GAME, NULL, level);

    /* ---- Check for earlier value ---- */
    
    l = T_to_list(pos);
    if (hash_test(TOADS_KEY, l))
    {
	list_free(l);
	g = hash_get_last();
	toads_trace(pos, g, "Seen it", level);
	return g;
    }

    /* ---- Check for Death Leap Principle ---- */

    i=2;
    thestate=Start;
    while ((thestate != Die) && (thestate != Live))
	thestate = nextstate[(int)thestate][(int)pos->board[i++]];

    if (thestate == Die)
    {
	hash_put(TOADS_KEY, l, zero);			/* memoize */
	toads_trace(pos, zero, "Death", level);
	return zero;
    }

    /* ---- Left's normal moves ---- */

    left = list_make();			/* left options */

    for (i = 2; i < pos->length+2; i++)
    {
	/* push */
	if ((pos->board[i] == Toad) && (pos->board[i+1] == Blank))
	{
	    pos->board[i] = Blank;
	    pos->board[i+1] = Toad;
	    list_insert(left, toads_position(pos, level+1));
	    pos->board[i] = Toad;
	    pos->board[i+1] = Blank;
	}

	/* jump */
	if ((pos->board[i] == Toad) && (pos->board[i+1] == Frog) &&
	    (pos->board[i+2] == Blank))
	{
	    pos->board[i] = Blank;
	    pos->board[i+2] = Toad;
	    list_insert(left, toads_position(pos, level+1));
	    pos->board[i] = Toad;
	    pos->board[i+2] = Blank;
	}
    }

    /* ---- Right's normal moves ---- */

    right = list_make();		/* right options */

    for (i = 2; i < pos->length+2; i++)
    {
	/* push */
	if ((pos->board[i] == Frog) && (pos->board[i-1] == Blank))
	{
	    pos->board[i] = Blank;
	    pos->board[i-1] = Frog;
	    list_insert(right, toads_position(pos, level+1));
	    pos->board[i] = Frog;
	    pos->board[i-1] = Blank;
	}

	/* jump */
	if ((pos->board[i] == Frog) && (pos->board[i-1] == Toad) &&
	    (pos->board[i-2] == Blank))
	{
	    moves = 1;
	    pos->board[i] = Blank;
	    pos->board[i-2] = Frog;
	    list_insert(right, toads_position(pos, level+1));
	    pos->board[i] = Frog;
	    pos->board[i-2] = Blank;
	}
    }

    g = make(left,right);
    hash_put(TOADS_KEY, l, g);			/* memoize */
    toads_trace(pos, g, NULL, level);
    
    return g;
}


/* ---------------------------------------------------------------------- *\
|  T_type toads_simplify(T_type pos)
|
|  Simplify a toads position as much as possible:
|  1. Replace dead pieces with walls.  A piece is dead if it cannot
|     move in the current game or any of its followers.
|  2. Eliminate "terminal toads" and "finished frogs", but record their
|     contribution to the game's value.
|  3. Replace dead spaces with walls.  A space is dead if no piece can
|     move to it in the current game or any of its followers.
|  4. Merge multiple walls.
|
|  For example:
|	tff--f--ttff--tfftt =(1)=>	|||--f--||||--tff||
|			      =(2)=>	||||----||||--tff|| - 2
|			      =(3)=>	||||||||||||--tff|| - 2
|  			      =(4)=>	--tff - 2
\* ---------------------------------------------------------------------- */
T_type toads_simplify(T_type pos)
{
    typedef enum {
	Start=0, GotWall, GotToad, GotToads, GotFrog, Done, Copy
    } State;

    static State newstate[5][4] = 		
    {
    /*   Wall     Toad      Frog     Blank */	/* State x T_Item -> State */

        {GotWall, GotToad,  Copy,    Copy},	/* Start    */
	{Done,    GotToads, Done,    Copy},	/* GotWall  */
	{Done,    GotToads, Copy,    Copy},	/* GotToad  */
	{Done,    GotToads, GotFrog, Copy},	/* GotToads */
	{Done,    GotToads, Done,    Copy}	/* GotFrog  */
    };

    State thestate;
    T_type New;
    int i, j, k, flag;

    New = T_new(pos->length);
    New->ttffbonus = pos->ttffbonus;

/*    toads_trace(pos, NO_GAME, "original position", 0); */

    /* ---- Cover dead toads and frogs with walls ---- */

    i = j = 2;
    thestate = GotWall;
    while (i < pos->length+2)
    {
	if (pos->board[i] == Blank)
	{
	    while (pos->board[i] == Blank)
	    {
		i++;
		New->board[j++] = Blank;
	    }
	    thestate = Start;
	}
	else if (thestate == Copy)	/* Did I copy last time? */
	    thestate = Start;
	else
	    thestate = GotWall;
	k=i;

	while ((i<pos->length+3) && (thestate != Done) && (thestate != Copy))
	    thestate = newstate[(int)thestate][(int)pos->board[i++]];

	if (thestate == Done)
	{
	    while (k<i)
	    {
		New->board[j++] = Wall;
		k++;
	    }
	}
	else			/* Copy or early abort */
	{
	    while (k<i)
		New->board[j++] = pos->board[k++];
	}
    }

/*    toads_trace(New, NO_GAME, "buried dead pieces", 0); */

    /* ---- Eliminate finished frogs ---- */

    for (i=j=k=flag=0; i <= pos->length; i++)
    {
	switch (New->board[i+2])
	{
	    case Wall:
		for ( ; k>0; k--)
		    New->board[(j++)+2] = Blank;
		for ( ; j<=i; j++)
		    New->board[j+2] = Wall;
		flag = 0;
		break;
	    case Frog:
		if (!flag)
		    New->ttffbonus -= k;
		else
		    New->board[(j++)+2] = New->board[i+2];
		break;
	    case Blank:
		if (!flag)
		    k++;
		else
		    New->board[(j++)+2] = New->board[i+2];
	        break;
	    case Toad:
		flag = 1;
		for ( ; k>0; k--)
		    New->board[(j++)+2] = Blank;
		New->board[(j++)+2] = New->board[i+2];
		break;
	    default:
		break;
	}
    }

/*    toads_trace(New, NO_GAME, "removed finished frogs", 0); */

    /* ---- Eliminate terminal toads ---- */

    for (i=j=New->length-1, k=flag=0; i>=-1; i--)
    {
	switch (New->board[i+2])
	{
	    case Wall:
		for ( ; k>0; k--)
		    New->board[(j--)+2] = Blank;
		for ( ; j>=i; j--)
		    New->board[j+2] = Wall;
		flag = 0;
		break;
	    case Toad:
		if (!flag)
		    New->ttffbonus += k;
		else
		    New->board[(j--)+2] = New->board[i+2];
		break;
	    case Blank:
		if (!flag)
		    k++;
		else
		    New->board[(j--)+2] = New->board[i+2];
	        break;
	    case Frog:
		flag = 1;
		for ( ; k>0; k--)
		    New->board[(j--)+2] = Blank;
		New->board[(j--)+2] = New->board[i+2];
		break;
	    default:
		break;
	}
    }

/*    toads_trace(New, NO_GAME, "removed terminal toads", 0); */

    /* ---- Mark blanks behind toads ---- */

    for (flag=i=0; i<New->length; i++)
    {
	switch (New->board[i+2])
	{
	    case Toad:
		flag = 1;
		break;
	    case Wall:
	        flag = 0;
		break;
	    case Blank:
	        if (!flag)
		    New->board[i+2] = Unknown;
	        break;
	    default:
		break;
	}
    }

    /* ---- Bury marks behind frogs, make others blanks again ---- */

    for (flag=0, i=New->length-1; i>=0; i--)
    {
	switch (New->board[i+2])
	{
	    case Frog:
		flag = 1;
		break;
	    case Wall:
	        flag = 0;
		break;
	    case Unknown:
		New->board[i+2] = (flag ? Blank : Wall);
	        break;
	    default:
		break;
	}
    }

/*    toads_trace(New, NO_GAME, "buried dead blanks", 0); */

    /* ---- Merge consecutive walls ---- */

    for (i=j=1; i < pos->length+3; i++)
    {
	if (New->board[i] == Wall)
	{
	    New->board[j++] = Wall;
	    New->components++;
	    while ((i < pos->length+2) && (New->board[i+1] == Wall))
		i++;
	}
	else
	    New->board[j++] = New->board[i];
    }
    New->board[j++] = Wall;
    New->length = j-4;
    New->components--;

/*    toads_trace(New, NO_GAME, "merged walls", 0); */

    return New;
}


/* ---------------------------------------------------------------------- *\
|  game_type toads_position(T_type pos, int level)
|
|  Evaluate a toads and frogs position.  Simplify the position, split
|  it into components, and add up their values, plus the ttff bonus.
\* ---------------------------------------------------------------------- */
game_type toads_position(T_type pos, int level)
{
    T_type comp;
    game_type g;
    int i, j, bonus;

    pos = toads_simplify(pos);
    bonus = pos->ttffbonus;
    pos->ttffbonus = 0;

    if (pos->components == 0)
    {
	g = zero;
	toads_trace(pos, zero, "Trivial!", level);
    }
    else if (pos->components == 1)
	g = toads_component(pos, level);
    else
    {
	toads_trace(pos, NO_GAME, NULL, level);

	g = zero;
	comp = T_new(pos->length);
	comp->board[0] = comp->board[1] = Wall;
	comp->components = 1;

	for (i=j=0; i<=pos->length; i++,j++)
	{
	    comp->board[j+2] = pos->board[i+2];

	    if (comp->board[j+2] == Wall)
	    {
		comp->board[j+3] = Wall;
		comp->length = j;
		g = plus(g, toads_component(comp, level+1));
		j=-1;
	    }
	}
	T_free(comp);
	toads_trace(pos, g, "Multiple components", level);
    }

    if (bonus != 0)
    {
	pos->ttffbonus = bonus;
	g = plus(g, num(Q_int(bonus)));
	toads_trace(pos, g, "TT/FF", level);
    }

    T_free(pos);
    return g;
}


/* ---------------------------------------------------------------------- *\
|  game_type toads(void)
|
|  Get a Toads and Frogs position and evaluate it.
\* ---------------------------------------------------------------------- */
game_type toads(void)
{
    game_type g;
    T_type pos;

    pos = input_toads();
    g = toads_position(pos, 0);
    T_free(pos);
    return g;
}

